/*
 * gimmix.c -- University of Applied Sciences Giessen
 *             MMIX implementation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

#include "gimmix-common.h"
#include "gimmix-arith.h"
#include "gimmix-io.h"
#include "gimmix-sim.h"
#include "abstime.h"


#define command_buf_size	1024	/* big, for floating point tests */


/**************************************************************/
/* ONLY FOR TESTING! */


int breakpoint;
int halted;
int resuming;
int interacting;
int interact_after_break;
int showing_stats;
octa g[1000];
int tracing_exceptions;
int stack_tracing;
int trace_threshold;
int gap;
int profile_gap;
int showing_source;
int profile_showing_source;
int buf_size;
int lring_size;

char *p;
int tracing;
int k;


int mmgetchars(char *buf, int size, octa addr, int stop) {
  return 0;
}


void mmputchars(unsigned char *buf, int size, octa addr) {
}


char stdin_chr(void) {
  return 'a';
}


/* ONLY FOR TESTING! */
/**************************************************************/


#define panic(m)	{ fprintf(stderr, "Panic: %s!\n", m); exit(-2); }


//<Preprocessor macros 11>
/* ATTENTION: We assume a standard C compiler. */
/* Then there is nothing to do here. */


//<Type declarations 9>
//typedef enum {
//  false, true
//} bool;

//typedef unsigned char byte;
//typedef unsigned int tetra;
//typedef struct {
//  tetra h;
//  tetra l;
//} octa;

/* simulated memory */

typedef struct {
  tetra tet;		/* the tetrabyte of simulated memory */
  tetra freq;		/* the number of times it was used as instruction */
  unsigned char bkpt;	/* breakpoint information for this tetrabyte */
  unsigned char file_no;	/* source file number, if known */
  unsigned short line_no;	/* source line number, if known */
} mem_tetra;

typedef struct mem_node_struct {
  octa loc;			/* location of the first of 512 tetrabytes */
  tetra stamp;				/* time stamp for treap balancing */
  struct mem_node_struct *left;		/* pointer to left subtree */
  struct mem_node_struct *right;	/* pointer to right subtree */
  mem_tetra dat[512];		/* the chunk of simulated tetrabytes */
} mem_node;

/* lists */

typedef enum {
  TRAP,    FCMP,    FUN,     FEQL,    FADD,    FIX,     FSUB,    FIXU,
  FLOT,    FLOTI,   FLOTU,   FLOTUI,  SFLOT,   SFLOTI,  SFLOTU,  SFLOTUI,
  FMUL,    FCMPE,   FUNE,    FEQLE,   FDIV,    FSQRT,   FREM,    FINT,
  MUL,     MULI,    MULU,    MULUI,   DIV,     DIVI,    DIVU,    DIVUI,
  ADD,     ADDI,    ADDU,    ADDUI,   SUB,     SUBI,    SUBU,    SUBUI,
  ADDU2,   ADDUI2,  ADDU4,   ADDUI4,  ADDU8,   ADDUI8,  ADDU16,  ADDUI16,
  CMP,     CMPI,    CMPU,    CMPUI,   NEG,     NEGI,    NEGU,    NEGUI,
  SL,      SLI,     SLU,     SLUI,    SR,      SRI,     SRU,     SRUI,
  BN,      BNB,     BZ,      BZB,     BP,      BPB,     BOD,     BODB,
  BNN,     BNNB,    BNZ,     BNZB,    BNP,     BNPB,    BEV,     BEVB,
  PBN,     PBNB,    PBZ,     PBZB,    PBP,     PBPB,    PBOD,    PBODB,
  PBNN,    PBNNB,   PBNZ,    PBNZB,   PBNP,    PBNPB,   PBEV,    PBEVB,
  CSN,     CSNI,    CSZ,     CSZI,    CSP,     CSPI,    CSOD,    CSODI,
  CSNN,    CSNNI,   CSNZ,    CSNZI,   CSNP,    CSNPI,   CSEV,    CSEVI,
  ZSN,     ZSNI,    ZSZ,     ZSZI,    ZSP,     ZSPI,    ZSOD,    ZSODI,
  ZSNN,    ZSNNI,   ZSNZ,    ZSNZI,   ZSNP,    ZSNPI,   ZSEV,    ZSEVI,
  LDB,     LDBI,    LDBU,    LDBUI,   LDW,     LDWI,    LDWU,    LDWUI,
  LDT,     LDTI,    LDTU,    LDTUI,   LDO,     LDOI,    LDOU,    LDOUI,
  LDSF,    LDSFI,   LDHT,    LDHTI,   CSWAP,   CSWAPI,  LDUNC,   LDUNCI,
  LDVTS,   LDVTSI,  PRELD,   PRELDI,  PREGO,   PREGOI,  GO,      GOI,
  STB,     STBI,    STBU,    STBUI,   STW,     STWI,    STWU,    STWUI,
  STT,     STTI,    STTU,    STTUI,   STO,     STOI,    STOU,    STOUI,
  STSF,    STSFI,   STHT,    STHTI,   STCO,    STCOI,   STUNC,   STUNCI,
  SYNCD,   SYNCDI,  PREST,   PRESTI,  SYNCID,  SYNCIDI, PUSHGO,  PUSHGOI,
  OR,      ORI,     ORN,     ORNI,    NOR,     NORI,    XOR,     XORI,
  AND,     ANDI,    ANDN,    ANDNI,   NAND,    NANDI,   NXOR,    NXORI,
  BDIF,    BDIFI,   WDIF,    WDIFI,   TDIF,    TDIFI,   ODIF,    ODIFI,
  MUX,     MUXI,    SADD,    SADDI,   MOR,     MORI,    MXOR,    MXORI,
  SETH,    SETMH,   SETML,   SETL,    INCH,    INCMH,   INCML,   INCL,
  ORH,     ORMH,    ORML,    ORL,     ANDNH,   ANDNMH,  ANDNML,  ANDNL,
  JMP,     JMPB,    PUSHJ,   PUSHJB,  GETA,    GETAB,   PUT,     PUTI,
  POP,     RESUME,  SAVE,    UNSAVE,  SYNC,    SWYM,    GET,     TRIP
} mmix_opcode;

typedef enum {
  rB,  rD,  rE,  rH, rJ,  rM,  rR,  rBB,
  rC,  rN,  rO,  rS, rI,  rT,  rTT, rK,
  rQ,  rU,  rV,  rG, rL,  rA,  rF,  rP,
  rW,  rX,  rY,  rZ, rWW, rXX, rYY, rZZ
} special_reg;


//<Global variables 19>

/* simulated memory */
tetra priority = 314159265;	/* pseudorandom time stamp counter */
mem_node *mem_root;		/* root of the treap */
mem_node *last_mem;		/* node most recently read or written */

/* object file loader */
FILE *mmo_file;			/* the input file */
int postamble;			/* have we encountered lop_post? */
int byte_count;			/* index of the next-to-be-read byte */
byte buf[4];			/* the most recently read bytes */
int yzbytes;			/* the two least significant bytes */
int delta;			/* difference for relative fixup */
tetra tet;			/* buf bytes packed big-endianwise */

/* names of special registers */
char *special_name[32] = {
  "rB",  "rD",  "rE",  "rH", "rJ",  "rM",  "rR",  "rBB",
  "rC",  "rN",  "rO",  "rS", "rI",  "rT",  "rTT", "rK",
  "rQ",  "rU",  "rV",  "rG", "rL",  "rA",  "rF",  "rP",
  "rW",  "rX",  "rY",  "rZ", "rWW", "rXX", "rYY", "rZZ"
};

/* */
char *myself;			/* argv[0], the name of this simulator */
char **cur_arg;			/* pointer to current argument */
bool interrupt;			/* has the user interrupted the simulation? */
bool profiling;			/* should we print the profile at the end? */
FILE *fake_stdin;		/* file substituted for the simulated StdIn */
FILE *dump_file;		/* file used for binary dumps */

char *usage_help[] = {
  " with these options: (<n>=decimal number, <x>=hex number)\n",
  "-t<n> trace each instruction the first n times\n",
  "-e<x> trace each instruction with an exception matching x\n",
  "-r    trace hidden details of the register stack\n",
  "-l<n> list source lines when tracing, filling gaps <= n\n",
  "-s    show statistics after each traced instruction\n",
  "-P    print a profile when simulation ends\n",
  "-L<n> list source lines with the profile\n",
  "-v    be verbose: show almost everything\n",
  "-q    be quiet: show only the simulated standard output\n",
  "-i    run interactively (prompt for online commands)\n",
  "-I    interact, but only after the program halts\n",
  "-b<n> change the buffer size for source lines\n",
  "-c<n> change the cyclic local register ring size\n",
  "-f<filename> use given file to simulate standard input\n",
  "-D<filename> dump a file for use by other simulators\n",
  ""
};

char *interactive_help[] = {
  "The interactive commands are:\n",
  "<return>  trace one instruction\n",
  "n         trace one instruction\n",
  "c         continue until halt or breakpoint\n",
  "q         quit the simulation\n",
  "s         show current statistics\n",
  "l<n><t>   set and/or show local register in format t\n",
  "g<n><t>   set and/or show global register in format t\n",
  "rA<t>     set and/or show register rA in format t\n",
  "$<n><t>   set and/or show dynamic register in format t\n",
  "M<x><t>   set and/or show memory octabyte in format t\n",
  "+<n><t>   set and/or show n additional octabytes in format t\n",
  " <t> is ! (decimal) or . (floating) or # (hex) or \" (string)\n",
  "     or <empty> (previous <t>) or =<value> (change value)\n",
  "@<x>      go to location x\n",
  "b[rwx]<x> set or reset breakpoint at location x\n",
  "t<x>      trace location x\n",
  "u<x>      untrace location <x>\n",
  "T         set current segment to Text_Segment\n",
  "D         set current segment to Data_Segment\n",
  "P         set current segment to Pool_Segment\n",
  "S         set current segment to Stack_Segment\n",
  "B         show all current breakpoints and tracepoints\n",
  "i<file>   insert commands from file\n",
  "-<option> change a tracing/listing/profile option\n",
  "-?        show the tracing/listing/profile options\n",
  ""
};

char command_buf[command_buf_size];
FILE *incl_file = NULL;		/* file of commands included by 'i' */
char cur_disp_mode = 'l';	/* 'l' or 'g' or '$' or 'M' */
char cur_disp_type = '!';	/* '!' or '.' or '#' or '"' */
bool cur_disp_set;		/* was the last <t> of the form =<val>? */
octa cur_disp_addr;		/* the h half is relevant only in mode 'M' */
octa cur_seg;			/* current segment offset */
char spec_reg_code[] = {
  rA,  rB,  rC,  rD,  rE,  rF,  rG,  rH,  rI,  rJ,  rK,  rL,  rM,
  rN,  rO,  rP,  rQ,  rR,  rS,  rT,  rU,  rV,  rW,  rX,  rY,  rZ
};
char spec_regg_code[] = {
  0,   rBB, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   rTT, 0,   0,   rWW, rXX, rYY, rZZ
};


/**************************************************************/


//<Subroutines 12>
/* ATTENTION: All "extern" declarations have been omitted. */


void print_hex(octa o) {
  if (o.h != 0) {
    printf("%x%08x", o.h, o.l);
  } else {
    printf("%x", o.l);
  }
}


void print_int(octa o) {
  tetra hi, lo, r, t;
  int j;
  char dig[20];

  hi = o.h;
  lo = o.l;
  if (lo == 0 && hi == 0) {
    printf("0");
  } else {
    if (hi & sign_bit) {
      printf("-");
      if (lo == 0) {
        hi = -hi;
      } else {
        lo = -lo;
        hi = ~hi;
      }
    }
    for (j = 0; hi; j++) {
      /* 64-bit division by 10 */
      r = ((hi % 10) << 16) + (lo >> 16);
      hi = hi / 10;
      t = ((r % 10) << 16) + (lo & 0xffff);
      lo = ((r / 10) << 16) + (t / 10);
      dig[j] = t % 10;
    }
    for (; lo; j++) {
      dig[j] = lo % 10;
      lo = lo / 10;
    }
    for (j--; j >= 0; j--) {
      printf("%c", dig[j] + '0');
    }
  }
}


/**************************************************************/

/* simulated memory */


mem_node *new_mem(void) {
  mem_node *p;

  p = (mem_node *) calloc(1, sizeof(mem_node));
  if (p == NULL) {
    panic("Can't allocate any more memory");
  }
  /* use Fibonacci hashing for stamp */
  p->stamp = priority;
  priority += 0x9e3779b9;  /* floor(2^32*(phi-1)) */
  return p;
}


mem_tetra *mem_find(octa addr) {
  octa key;
  int offset;
  mem_node *p;
  mem_node **q;
  mem_node **l, **r;

  p = last_mem;
  key.h = addr.h;
  key.l = addr.l & 0xfffff800;
  offset = addr.l & 0x7fc;
  if (p->loc.l != key.l || p->loc.h != key.h) {
    //<Search for key in the treap, setting last_mem and p to its location 21>
    for (p = mem_root; p; ) {
      if (key.l == p->loc.l && key.h == p->loc.h) {
        /* tetrabyte found in this node */
        last_mem = p;
        return &p->dat[offset >> 2];
      }
      if ((key.l < p->loc.l && key.h <= p->loc.h) || key.h < p->loc.h) {
        p = p->left;
      } else {
        p = p->right;
      }
    }
    /* the tetrabyte is not yet in the treap, must insert a new node */
    for (p = mem_root, q = &mem_root;
         p != NULL && p->stamp < priority;
         p = *q) {
      if ((key.l < p->loc.l && key.h <= p->loc.h) || key.h < p->loc.h) {
        q = &p->left;
      } else {
        q = &p->right;
      }
    }
    *q = new_mem();
    (*q)->loc = key;
    //<Fix up the subtrees of *q 22>
    l = &(*q)->left;
    r = &(*q)->right;
    while (p) {
      if ((key.l < p->loc.l && key.h <= p->loc.h) || key.h < p->loc.h) {
        *r = p;
        r = &p->left;
        p = *r;
      } else {
        *l = p;
        l = &p->right;
        p = *l;
      }
    }
    *l = NULL;
    *r = NULL;
    /* set p and last_mem */
    p = *q;
    last_mem = p;
  }
  return &p->dat[offset >> 2];
}


/**************************************************************/
/* ONLY FOR TESTING! */


void TEST(void) {
  octa o;
//  o.h = 0x12345678;
//  o.l = 0x9abcdef0;
//  o.h = 0;
//  o.l = 0x12345678;
//  print_hex(o);
//  printf("\n");
//  print_int(o);
//  printf("\n");
  for (o.l = 0; o.l < 0x8000; o.l++) {
    o.h = 0x40000000;
    print_hex(o);
    printf("\n");
    mem_find(o);
  }
  panic("The excution arrived at end of TEST() and stops now");
}


void show_stats(bool t) {
}


/* ONLY FOR TESTING! */
/**************************************************************/


#define mmo_file_name		(*cur_arg)


void scan_option(char *arg, bool usage) {
  int k;

  switch (*arg) {
    case 't':
      if (strlen(arg) > 10) {
        trace_threshold = 0xffffffff;
      } else {
        if (sscanf(arg + 1, "%d", &trace_threshold) != 1) {
          trace_threshold = 0;
        }
      }
      return;
    case 'e':
      if (*(arg + 1) == '\0') {
        tracing_exceptions = 0xff;
      } else {
        if (sscanf(arg + 1, "%x", &tracing_exceptions) != 1) {
          tracing_exceptions = 0;
        }
      }
      return;
    case 'r':
      stack_tracing = true;
      return;
    case 's':
      showing_stats = true;
      return;
    case 'l':
      if (*(arg + 1) == '\0') {
        gap = 3;
      } else {
        if (sscanf(arg + 1, "%d", &gap) != 1) {
          gap = 0;
        }
      }
      showing_source = true;
      return;
    case 'L':
      if (*(arg + 1) == '\0') {
        profile_gap = 3;
      } else {
        if (sscanf(arg + 1, "%d", &profile_gap) != 1) {
          profile_gap = 0;
        }
      }
      profile_showing_source = true;
      profiling = true;
      return;
    case 'P':
      profiling = true;
      return;
    case 'v':
      trace_threshold = 0xffffffff;
      tracing_exceptions = 0xff;
      stack_tracing = true;
      showing_stats = true;
      gap = 10;
      showing_source = true;
      profile_gap = 10;
      profile_showing_source = true;
      profiling = true;
      return;
    case 'q':
      trace_threshold = 0;
      tracing_exceptions = 0;
      stack_tracing = false;
      showing_stats = false;
      showing_source = false;
      profile_showing_source = false;
      profiling = false;
      return;
    case 'i':
      interacting = true;
      return;
    case 'I':
      interact_after_break = true;
      return;
    case 'b':
      if (sscanf(arg + 1, "%d", &buf_size) != 1) {
        buf_size = 0;
      }
      return;
    case 'c':
      if (sscanf(arg + 1, "%d", &lring_size) != 1) {
        lring_size = 0;
      }
      return;
    case 'f':
      //<Open a file for simulated standard input 145>
      if (fake_stdin != NULL) {
        fclose(fake_stdin);
      }
      fake_stdin = fopen(arg + 1, "r");
      if (fake_stdin == NULL) {
        fprintf(stderr, "Sorry, I can't open file %s!\n", arg + 1);
      } else {
        mmix_fake_stdin(fake_stdin);
      }
      return;
    case 'D':
      //<Open a file for dumping binary output 146>
      dump_file = fopen(arg + 1, "wb");
      if (dump_file == NULL) {
        fprintf(stderr, "Sorry, I can't open file %s!\n", arg + 1);
      }
      return;
    default:
      if (usage) {
        fprintf(stderr,
                "Usage: %s <options> progfile command-line-args...\n",
                myself);
        for (k = 0; usage_help[k][0]; k++) {
          fprintf(stderr, usage_help[k]);
        }
        exit(-1);
      } else {
        for (k = 0; usage_help[k][1] != 'b'; k++) {
          printf(usage_help[k]);
        }
      }
      return;
  }
}


void catchint(int n) {
  interrupt = true;
  signal(SIGINT, catchint);
}


int main(int argc, char *argv[]) {
  char *alt_name;
  int repeating;
  bool ready;

  //<Local registers 62>
  mmix_io_init();
  //<Process the command line 142>
  myself = argv[0];
  for (cur_arg = argv + 1;
       *cur_arg != NULL && (*cur_arg)[0] == '-';
       cur_arg++) {
    scan_option(*cur_arg + 1, true);
  }
  if (*cur_arg == NULL) {
    /* exit with usage note */
    scan_option("?", true);
  }
  /* set the argc of the user program */
  argc -= cur_arg - argv;
  //<Initialize everything 14>
  /* do a quick check to see if arithmetic is in trouble */
  if (shift_left(neg_one, 1).h != 0xffffffff) {
    panic("Incorrect implementation of type tetra");
  }
  /* allocate initial chunk of memory for pool segment */
  mem_root = new_mem();
  mem_root->loc.h = 0x40000000;
  last_mem = mem_root;
  /* open MMO file */
  mmo_file = fopen(mmo_file_name, "rb");
  if (mmo_file == NULL) {
    alt_name = (char *) calloc(strlen(mmo_file_name) + 5, sizeof(char));
    if (alt_name == NULL) {
      panic("Can't allocate space for alternate MMO file name");
    }
    sprintf(alt_name, "%s.mmo", mmo_file_name);
    mmo_file = fopen(alt_name, "rb");
    if (mmo_file == NULL) {
      fprintf(stderr,
              "Can't open the object file %s or %s!\n",
              mmo_file_name, alt_name);
      exit(-3);
    }
    free(alt_name);
  }
  byte_count = 0;
  /* install signal handler */
  signal(SIGINT, catchint);
  //<Load the command line arguments 163>
  //<Get ready to UNSAVE the initial context 164>
  while (1) {
    if (interrupt && !breakpoint) {
      breakpoint = true;
      interacting = true;
      interrupt = false;
    } else {
      breakpoint = false;
      if (interacting) {
        //<Interact with the user 149>
interact:
        //<Put a new command in command_buf 150>
        ready = false;
incl_read:
        while (incl_file != NULL && !ready) {
          if (fgets(command_buf, command_buf_size, incl_file) == NULL) {
            fclose(incl_file);
            incl_file = NULL;
          } else {
            if (command_buf[0] != '\n' &&
                command_buf[0] != 'i' &&
                command_buf[0] != '%') {
              if (command_buf[0] == ' ') {
                printf(command_buf);
              } else {
                ready = true;
              }
            }
          }
        }
        while (!ready) {
          printf("mmix> ");
          fflush(stdout);
          if (fgets(command_buf, command_buf_size, stdin) == NULL) {
            command_buf[0] = 'q';
          }
          if (command_buf[0] != 'i') {
            ready = true;
          } else {
            command_buf[strlen(command_buf) - 1] = '\0';
            incl_file = fopen(command_buf + 1, "r");
            if (incl_file != NULL) {
              goto incl_read;
            }
            if (isspace(command_buf[1])) {
              incl_file = fopen(command_buf + 2, "r");
            }
            if (incl_file != NULL) {
              goto incl_read;
            }
            printf("Can't open file '%s'!\n", command_buf + 1);
          }
        }
        /* analyze & execute command */
        p = command_buf;
        repeating = 0;
        switch (*p) {
          case '\n':
          case 'n':
            breakpoint = true;
            tracing = true;
            goto resume_simulation;
          case 'c':
            goto resume_simulation;
          case 'q':
            goto end_simulation;
          case 's':
            show_stats(true);
            goto interact;
          case '-':
            k = strlen(p);
            if (p[k - 1] == '\n') {
              p[k - 1] = '\0';
            }
            scan_option(p + 1, false);
            goto interact;
          //<Cases that change cur_disp_mode 152>
          case 'l':
          case 'g':
          case '$':
            cur_disp_mode = *p++;
            for (cur_disp_addr.l = 0; isdigit(*p); p++) {
              cur_disp_addr.l = 10 * cur_disp_addr.l + *p - '0';
            }
            goto new_mode;
          case 'r':
            p++;
            cur_disp_mode = 'g';
            if (*p < 'A' || *p > 'Z') {
              goto what_say;
            }
            if (*(p + 1) != *p) {
              cur_disp_addr.l = spec_reg_code[*p - 'A'];
              p++;
            } else
            if (spec_regg_code[*p - 'A'] != 0) {
              cur_disp_addr.l = spec_regg_code[*p - 'A'];
              p += 2;
            } else {
              goto what_say;
            }
            goto new_mode;
          case 'M':
            cur_disp_mode = *p;
            cur_disp_addr = scan_hex(p + 1, cur_seg);
            cur_disp_addr.l &= -8;
            p = next_char;
new_mode:
            cur_disp_set = false;  /* the '=' is remembered only by '+' */
            repeating = 1;
            goto scan_type;
          case '+':
            if (!isdigit(*(p + 1))) {
              repeating = 1;
            }
            for (p++; isdigit(*p); p++) {
              repating = 10 * repeating + *p - '0';
            }
            if (repeating) {
              if (cur_disp_mode == 'M') {
                cur_disp_addr = incr(cur_disp_addr, 8);
              } else {
                cur_disp_addr.l++;
              }
            }
            goto scan_type;
          //<Cases that define cur_disp_type 153>
          case '!':
          case '.':
          case '#':
          case '"':
            cur_disp_set = false;
            repeating = 1;
set_type:
            cur_disp_type = *p++;
            break;
scan_type:
            if (*p == '!' || *p == '.' || *p == '#' || *p == '"') {
              goto set_type;
            }
            if (*p != '=') {
              break;
            }
            goto scan_eql;
          case '=':
            repeating = 1;
scan_eql:
            cur_disp_set = true;
            val = zero_octa;
            if (*++p == '#') {
              cur_disp_type = *p;
              val = scan_hex(p + 1, zero_octa);
            } else
            if (*p == '"' || *p == '\'') {
              goto scan_string;
            } else {
              cur_disp_type = (scan_const(p) > 0 ? '.' : '!');
            }
            p = next_char;
            if (*p != ',') {
              break;
            }
            val.h = 0;
            val.l &= 0xff;
scan_string:
            cur_disp_type = '"';
            //<Scan a string constant 155>
            break;
          //<Cases that set and clear tracing and breakpoints 161>
          default:
what_say:
            k = strlen(command_buf);
            if (k < 10 && command_buf[k - 1] == '\n') {
              command_buf[k - 1] = '\0';
            } else {
              strcpy(command_buf + 9, "...");
            }
            printf("Eh? Sorry, I don't understand '%s'. (Type h for help)\n",
                   command_buf);
            goto interact;
          case 'h':
            for (k = 0; interactive_help[k][0]; k++) {
              printf(interactive_help[k]);
            }
            goto interact;
        }
check_syntax:
        if (*p != '\n') {
          if (*p == '\0') {
incomplete_str:
            printf("Syntax error: Incomplete command!\n");
          } else {
            p[strlen(p) - 1] = '\0';
            printf("Syntax error: I'm ignoring '%s'!\n", p);
          }
        }
        while (repeating) {
          //<Display and/or set the value of the current octabyte 156>
        }
        goto interact;
resume_simulation: ;
      }
    }
    if (halted) {
      break;
    }
    do {
      //<Perform one instruction 60>
    } while ((!interrupt && !breakpoint) || resuming);
    if (interact_after_break) {
      interacting = true;
      interact_after_break = false;
    }
  }
end_simulation:
  if (profiling) {
    //<Print all the frequency counts 53>
  }
  if (interacting || profiling || showing_stats) {
    show_stats(true);
  }
  /* provide rudimentary feedback for non-interactive runs */
  return g[255].l;
}
